// FLAC helper library
// Copyright (C) 2003 Janne Hyvrinen
//
// Changes:
//  0.0.1 (2003-05-24): created

#ifndef _flac_lib_helper_h
#define _flac_lib_helper_h

#define FL_READER   // uncomment to provide own reader

#include <ogg/ogg.h>
#include "FLAC/all.h"
#include "OggFLAC/all.h"
#include "stdlib.h"
#include "string.h"

#ifdef FL_READER
#include <windows.h>
#endif

// -------------------------------------

class Reader {
//protected:
//    bool aborting, prebuffering;

public:
    Reader() { /*aborting = 0; prebuffering = 0;*/ }
    virtual ~Reader() {};
    virtual int read ( void *ptr, int size ) = 0;
    virtual int write ( void *ptr, int size ) = 0;
    virtual __int64 seek ( __int64 offset ) = 0;
    virtual __int64 get_position() = 0;
    virtual __int64 get_length() = 0;
    virtual int set_eof ( __int64 ofs ) { return 0; }
    virtual int can_seek() { return 0; }
    //virtual int canwrite() { return 0; }
    //virtual void setabort ( bool a ) { aborting = a; }
    //virtual void setprebuf ( bool p ) { prebuffering = p; }
    //virtual void do_prebuf() {}
    //virtual void idle() { Sleep (10); }
};

class flac_decoder {
public:
    static int is_flac ( Reader *r );
    static int is_ogg_flac ( Reader *r );

protected:
    //Reader *pReader;
    FLAC__SeekableStreamDecoder *seekable_stream_decoder;
    FLAC__StreamDecoder *stream_decoder;
    OggFLAC__StreamDecoder *ogg_stream_decoder;

    FLAC__int32 *reservoir;
    unsigned samples_in_reservoir;
    int reservoir_size;
    int abort_flag, eof;

public:
    Reader *pReader;
    FLAC__StreamMetadata streaminfo;
    FLAC__bool status_ok;

public:
    flac_decoder ( Reader *r );
    ~flac_decoder();

    int decode ( FLAC__byte *output, unsigned int samples, int decode_to_float );
    void cleanup();

private:
    int pack_pcm ( FLAC__byte *data, FLAC__int32 *input, unsigned wide_samples, unsigned channels, unsigned bps );
    int convert_to_float ( float *data, FLAC__int32 *input, unsigned wide_samples, unsigned channels, unsigned bps );

    static FLAC__SeekableStreamDecoderReadStatus readCallback_ ( const FLAC__SeekableStreamDecoder *decoder, FLAC__byte buffer[], unsigned *bytes, void *client_data );
    static FLAC__SeekableStreamDecoderSeekStatus seekCallback_ ( const FLAC__SeekableStreamDecoder *decoder, FLAC__uint64 absolute_byte_offset, void *client_data );
    static FLAC__SeekableStreamDecoderTellStatus tellCallback_ ( const FLAC__SeekableStreamDecoder *decoder, FLAC__uint64 *absolute_byte_offset, void *client_data );
    static FLAC__SeekableStreamDecoderLengthStatus lengthCallback_ ( const FLAC__SeekableStreamDecoder *decoder, FLAC__uint64 *stream_length, void *client_data );
    static FLAC__bool eofCallback_ ( const FLAC__SeekableStreamDecoder *decoder, void *client_data );
    static FLAC__StreamDecoderWriteStatus writeCallback_ ( const FLAC__SeekableStreamDecoder *decoder, const FLAC__Frame *frame, const FLAC__int32 * const buffer[], void *client_data );
    static void metadataCallback_ ( const FLAC__SeekableStreamDecoder *decoder, const FLAC__StreamMetadata *metadata, void *client_data );
    static void errorCallback_ ( const FLAC__SeekableStreamDecoder *decoder, FLAC__StreamDecoderErrorStatus status, void *client_data );

    static FLAC__StreamDecoderReadStatus readCallback_ns ( const FLAC__StreamDecoder *decoder, FLAC__byte buffer[], unsigned *bytes, void *client_data );
    static FLAC__StreamDecoderWriteStatus writeCallback_ns ( const FLAC__StreamDecoder *decoder, const FLAC__Frame *frame, const FLAC__int32 * const buffer[], void *client_data );
    static void metadataCallback_ns ( const FLAC__StreamDecoder *decoder, const FLAC__StreamMetadata *metadata, void *client_data );
    static void errorCallback_ns ( const FLAC__StreamDecoder *decoder, FLAC__StreamDecoderErrorStatus status, void *client_data );

    static FLAC__StreamDecoderReadStatus readCallback_ogg_ns ( const OggFLAC__StreamDecoder *decoder, FLAC__byte buffer[], unsigned *bytes, void *client_data );
    static FLAC__StreamDecoderWriteStatus writeCallback_ogg_ns ( const OggFLAC__StreamDecoder *decoder, const FLAC__Frame *frame, const FLAC__int32 * const buffer[], void *client_data );
    static void metadataCallback_ogg_ns ( const OggFLAC__StreamDecoder *decoder, const FLAC__StreamMetadata *metadata, void *client_data );
    static void errorCallback_ogg_ns ( const OggFLAC__StreamDecoder *decoder, FLAC__StreamDecoderErrorStatus status, void *client_data );
};

class flac_encoder {
public:
    typedef struct {
        unsigned int channels;
        unsigned int bits_per_sample;
        unsigned int sample_rate;
        FLAC__uint64 total_samples;
        FLAC__bool   create_ogg;
        FLAC__bool   verify;
        FLAC__bool   streamable_subset;

        FLAC__bool   do_exhaustive_model_search;
        FLAC__bool   do_escape_coding;
        FLAC__bool   do_mid_side_stereo;
        FLAC__bool   loose_mid_side_stereo;
        unsigned int qlp_coeff_precision;
        unsigned int min_residual_partition_order;
        unsigned int max_residual_partition_order;
        unsigned int rice_parameter_search_dist;
        unsigned int max_lpc_order;
        unsigned int blocksize;
    } flac_encoder_parameters;

    static int compression_level ( flac_encoder_parameters &parameters, int level );

protected:
    //Reader *pReader;
    FLAC__SeekableStreamEncoder *seekable_stream_encoder;
    OggFLAC__StreamEncoder *ogg_stream_encoder;

    unsigned int channels;
    unsigned int bits_per_sample;

    FLAC__int32 *encode_buffer;

public:
    Reader *pReader;
    FLAC__StreamMetadata **metadata;
    int metadata_blocks;
    FLAC__bool status_ok;
    char error_message[512];

public:
    flac_encoder ( Reader *r, flac_encoder_parameters &parameters, FLAC__StreamMetadata *comments );
    ~flac_encoder();

    int encode ( const void *buffer, unsigned int samples, int source_bps=0, double float_scale=32768. );
    void cleanup();

private:
    int get_state();
    int FLAC_pcm ( const FLAC__byte *data, FLAC__int32 *output, unsigned wide_samples );
    int FLAC_float32 ( const float *data, FLAC__int32 *output, unsigned wide_samples, double scale );
    int FLAC_float64 ( const double *data, FLAC__int32 *output, unsigned wide_samples, double scale );

    static FLAC__StreamEncoderWriteStatus writeCallback_ ( const FLAC__SeekableStreamEncoder *encoder, const FLAC__byte buffer[], unsigned bytes, unsigned samples, unsigned current_frame, void *client_data );
    static FLAC__SeekableStreamEncoderSeekStatus seekCallback_ ( const FLAC__SeekableStreamEncoder *encoder, FLAC__uint64 absolute_byte_offset, void *client_data );

    static FLAC__StreamEncoderWriteStatus writeCallback_ogg_ns ( const OggFLAC__StreamEncoder *encoder, const FLAC__byte buffer[], unsigned bytes, unsigned samples, unsigned current_frame, void *client_data );
};

#ifdef FL_READER

class Reader_file : public Reader {
private:
    HANDLE fh;

public:
    Reader_file ( HANDLE _fh ) {
        fh = _fh;
    }

    ~Reader_file() {
        if ( fh != INVALID_HANDLE_VALUE ) {
            CloseHandle ( fh );
            fh = INVALID_HANDLE_VALUE;
        }
    }

    virtual int read ( void *ptr, int size ) {
        DWORD bytesread;
        if ( !ReadFile (fh, ptr, size, &bytesread, NULL) ) return 0;
        return bytesread;
    }

    virtual int write ( void *ptr, int size ) {
        DWORD byteswritten;
        if ( !WriteFile (fh, ptr, size, &byteswritten, NULL) ) return 0;
        return byteswritten;
    }

    virtual __int64 seek ( __int64 offset ) {
        LONG high = (DWORD)(offset >> 32);
        DWORD low = SetFilePointer ( fh, (LONG)offset, &high, FILE_BEGIN );
        if ( (low == 0xFFFFFFFF) && (GetLastError() != NO_ERROR) ) return -1;
        return ((__int64)high << 32) | low;
    }

    virtual __int64 get_position() {
        LONG high = 0;
        DWORD low = SetFilePointer ( fh, 0, &high, FILE_CURRENT );
        if ( (low == 0xFFFFFFFF) && (GetLastError() != NO_ERROR) ) return -1;
        return ((__int64)high << 32) | low;
    }

    virtual __int64 get_length() {
        ULONG high;
        DWORD low = GetFileSize ( fh, &high );
        if ( (low == INVALID_FILE_SIZE) && (GetLastError() != NO_ERROR) ) return -1;
        return ((__int64)high << 32) | low;
    }

    virtual int set_eof ( __int64 offset ) {
        if ( seek (offset) != offset ) return 0;
        return SetEndOfFile ( fh );
    };

    virtual int can_seek () {
        if ( GetFileType (fh) == FILE_TYPE_DISK ) return 1;
        return 0;
    }
};

#endif // FL_READER

#endif // _flac_lib_helper_h
